import argparse
import importlib
import os
import pkgutil
from importlib.metadata import version

from utils import log_utils


class Cli:
    def __init__(self):
        self.parser = argparse.ArgumentParser(
            prog="ci-cerberus",
            description=(
                f"      🐕 ci-cerberus v{version('ci-cerberus')}\n\n"
                "       Scans GitHub workflows for known vulnerable actions using the NIST National Vulnerability Database (NVD) API\n\n"  # noqa: E501
                "       More information on the NVD can be found at https://nvd.nist.gov/developers/vulnerabilities\n"  # noqa: E501
                "       Source code can be found at https://github.com/gavinroderick/ci-cerberus"
            ),
            usage="%(prog)s [options]",
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )

        self._add_help()
        self._add_debug()
        self._add_version()

        # Add subparsers & specify folder location
        self.subparsers = self.parser.add_subparsers(title="commands", metavar="")
        self._add_commands()

    def run(self, args=None):
        parsed_args = self.parser.parse_args(args)

        # Setup logging globally (cli)
        log_utils.setup_logger(parsed_args.debug)

        if hasattr(parsed_args, "version"):
            self.parser.print_version()
            return
        elif hasattr(parsed_args, "command"):
            parsed_args.command(parsed_args)
        else:
            self.parser.print_help()

    def _add_help(self):
        self.parser.add_argument(
            "-g",
            "--directory",
            type=str,
            metavar="",
            default=".",
            help="Directory where GitHub workflows are located (relative to current directory)",
        )

    def _add_debug(self):
        self.parser.add_argument(
            "-d",
            "--debug",
            action="store_true",
            help="Enable debug logging output (see what's happening under the hood!)",
        )

    def _add_version(self):
        self.parser.add_argument(
            "-v",
            "--version",
            action="version",
            version=f"{version("ci-cerberus")}",
            help="Display the version of the application",
        )

    def _add_commands(self):
        commands_path = os.path.join(os.path.dirname(__file__), "commands")

        for _, name, _ in pkgutil.iter_modules([commands_path]):
            module = importlib.import_module(f"cli.commands.{name}")

            if hasattr(module, "register_command"):
                module.register_command(self.subparsers)


def main():
    cli_instance = Cli()
    cli_instance.run()



import logging


def register_command(subparsers):
    parser = subparsers.add_parser(
        "test", help="test command to verify the application is functioning"
    )
    parser.set_defaults(command=handle_test)


def handle_test(args):
    logger = logging.getLogger(__name__)
    logger.info("You called the test command! Everything's working fine")
    logger.debug("(and you have verbose mode turned on)")



import logging
import os.path
from typing import Iterator, List, Optional

import yaml

from githubparsers.models.job import Job
from githubparsers.models.step import Step
from githubparsers.models.workflow import Workflow

logger = logging.getLogger(__name__)


class WorkflowParser:
    def __init__(self, github_directory: str = ".github/workflows"):
        self.github_directory = os.path.normpath(github_directory)

    def get_workflows(self) -> Optional[List[Workflow]]:
        files = self._find_workflow_files()
        if not files:
            return None

        workflows = []
        for file in files:
            workflow = self._parse_yaml(file)
            if not workflow:
                logger.error(f"❌ unable to parse workflow from {file}")
            workflows.append(workflow)

        if len(workflows) > 0:
            for workflow in workflows:

                steps = []
                for job in workflow.jobs:
                    for step in job.steps:
                        steps.append(step)

                logger.info(
                    f"      🎯 found {len(steps)} step{'s' if len(steps) > 1 else ''} in `{workflow.file_name or workflow.name}`"
                )

                for step in steps:
                    logger.info(
                        f"          - `{step.uses}` (step: `{step.id or step.name or 'No name'}`)"
                    )
                logger.info("")

        return workflows

    def _find_workflow_files(self) -> List[str] | None:
        absolute_path = os.path.abspath(self.github_directory)
        logger.debug(f"🔎 looking for workflow files in {absolute_path}")

        try:
            file_names = os.listdir(self.github_directory)
            if not file_names:
                logger.warning(f"❓couldn't find any files in {absolute_path}")
                return None

            files = []
            for file in file_names:
                files.append(os.path.join(self.github_directory, file))

            logger.info(f"🎯 found {len(file_names)} workflows in {absolute_path}")
            if files:
                for file in file_names:
                    logger.info(f"    - {file} ")
                logger.info("")

            return files
        except FileNotFoundError:
            logger.error(f"❓ no directory found at {absolute_path}")
            return None

    def _parse_yaml(self, file_path: str) -> Workflow:
        with open(file_path, "r") as file:
            logger.debug(f"📄 parsing workflow file {file.name}")
            workflow_yaml = yaml.load(file.read(), Loader=yaml.SafeLoader)
            return self._parse_workflow(workflow_yaml, file_path)

    def _parse_workflow(self, workflow_yaml, file_path) -> Optional[Workflow]:
        jobs = list(self._parse_jobs(workflow_yaml.get("jobs")))
        if not jobs:
            logger.debug(f"❓ couldn't find any jobs in {workflow_yaml}")

        return Workflow(
            id=workflow_yaml.get("id"),
            name=workflow_yaml.get("name"),
            jobs=jobs,
            file_name=file_path,
        )

    def _parse_jobs(self, jobs_yaml) -> Iterator[Job]:
        for key, job in jobs_yaml.items():
            steps = list(self._parse_steps(job.get("steps")))

            if not steps:
                logger.debug("❌ couldn't find any steps in job - skipping")

            yield Job(
                key=key,
                name=job.get("name"),
                runs_on=job.get("runs-on"),
                steps=steps,
            )

    @staticmethod
    def _parse_steps(steps_yaml) -> Iterator[Step]:
        # logger.debug(f"found step(s):")
        # logger.debug(yaml.dump(steps_yaml))
        for step in steps_yaml:
            uses = step.get("uses").split("@")[0]

            yield Step(
                id=step.get("id"),
                name=step.get("name"),
                uses=uses,
                with_arguments=step.get("with"),
            )



import logging
import sys


def setup_logger(verbose: bool = False) -> None:
    if verbose:
        logging.basicConfig(
            level="DEBUG",
            format="%(asctime)s [%(levelname)s] %(name)s: %(message)s",
            handlers=[logging.StreamHandler(sys.stdout)],
        )
    else:
        logging.basicConfig(
            level="INFO",
            format="%(message)s",
            handlers=[logging.StreamHandler(sys.stdout)],
        )



import logging
from dataclasses import dataclass, field
from typing import Any, Dict, List

from models.cve import Cve
from models.cvss import Cvss

logger = logging.getLogger(__name__)


@dataclass
class CvesResponse:
    """
    DTO for response from NVD CVEs endpoint.

    Contains a list of CVEs, number of results returned & time data was retrieved.
    """

    total_results: int
    timestamp: str
    cves: List[Cve] = field(default_factory=list)

    @staticmethod
    def parse_json(data: Dict[str, Any]) -> "CvesResponse":
        """Helper method to parse CVE from raw JSON into CvesResponse"""
        logger.debug("DEBUG: parsing CVE response")
        cves = []

        for item in data.get("vulnerabilities", []):
            cve_data = item.get("cve", {})
            cve_id = cve_data.get("id")

            descriptions = {}
            for description in cve_data.get("descriptions", []):
                descriptions[description.get("lang")] = description["value"]

            description_english = descriptions.get(
                "en", "Couldn't find an English description for this CVE"
            )

            cvss_scores = []
            for metric in cve_data.get("metrics", {}).get("cvssMetricV31", []):
                cvss_data = metric.get("cvssData", {})
                cvss = Cvss(
                    source=metric.get("source"),
                    type=metric.get("type"),
                    base_score=cvss_data.get("baseScore"),
                    severity=cvss_data.get("baseSeverity"),
                    vector=cvss_data.get("vectorString"),
                )
                cvss_scores.append(cvss)

            weaknesses = []
            for weakness in cve_data.get("weaknesses", []):
                for description in weakness.get("description", []):
                    if description.get("lang") == "en":
                        weaknesses.append(description["value"])

            cpes = []
            for config in cve_data.get("configurations", []):
                for node in config.get("nodes", []):
                    for match in node.get("cpeMatch", []):
                        if match.get("vulnerable"):
                            cpes.append(match.get("criteria"))

            references = []
            for reference in cve_data.get("references", []):
                reference_url = reference.get("url")
                references.append(reference_url)

            cves.append(
                Cve(
                    cve_id=cve_id,
                    description=description_english,
                    cvss_scores=cvss_scores,
                    weaknesses=weaknesses,
                    cpes=cpes,
                    references=references,
                )
            )

        return CvesResponse(
            total_results=data.get("totalResults"),
            timestamp=data.get("timestamp"),
            cves=cves,
        )



import logging
from typing import Any, Dict, Optional
from urllib.parse import urlencode

import requests

logger = logging.getLogger(__name__)


class BaseHttpClient:
    def __init__(self, base_url: str, headers: Optional[Dict[str, str]] = None) -> None:
        self.base_url = base_url.rstrip("/")
        self.headers = headers or {}

    def get(self, path: str, params: Optional[Dict[str, str]]) -> Optional[Any]:
        logger.debug(f"called BaseHttpClient.get({path}, {params})")
        url = f"{self.base_url}/{path}"
        if params:
            url = f"{url}?{urlencode(params)}"

        try:
            logger.debug(f"sending request to {url}")
            response = requests.get(url, headers=self.headers)
            response.raise_for_status()
            return response.json()
        except requests.HTTPError as http_error:
            logger.error(
                f"HTTPError occurred making GET request to {url}: {http_error}"
            )
        except requests.RequestException as request_exception:
            logger.error(
                f"RequestException occurred making GET request to {url}: {request_exception}"
            )

    def post(self, path: str, body: Optional[Any]) -> requests.Response:
        raise NotImplementedError()



from dataclasses import dataclass
from typing import List

from models.cvss import Cvss


@dataclass
class Cve:
    cve_id: str
    description: str
    cvss_scores: List[Cvss]
    weaknesses: List[str]
    cpes: List[str]
    references: List[str]



import logging

from githubparsers import workflowparser
from httpclients.nvd_httpclient import NvdHttpClient
from importlib.metadata import version
from typing import Tuple

from models.cve import Cve


def register_command(subparsers):
    parser = subparsers.add_parser(
        "scan", help="Scan workflow files for vulnerabilities"
    )
    parser.set_defaults(command=handle_scan)


def handle_scan(action_name: str):
    logger = logging.getLogger(__name__)

    logger.info(f"🐕 starting ci-cerberus! (v{version('ci-cerberus')})\n")
    logger.debug("(debug mode enabled)")

    parser = workflowparser.WorkflowParser()
    workflows = parser.get_workflows()
    steps_data = _get_steps(workflows)

    nvd_client = NvdHttpClient("https://services.nvd.nist.gov/")

    steps_with_cves: dict[(str, str), list[Cve]] = {}

    for step_obj, workflow_file_name in steps_data:
        cves_response = nvd_client.find_cves_by_keywords(step_obj.uses)
        if cves_response.cves:
            steps_with_cves[(step_obj.uses, workflow_file_name)] = cves_response.cves

    _log_cve_details(steps_with_cves, logger)


def _log_cve_details(
    steps_with_cves: dict[(str, str), list[Cve]], logger: logging.Logger
):
    logger.info(f"🔍 found {len(steps_with_cves)} steps with potential CVEs")
    for (step, workflow_file_name), cves in steps_with_cves.items():
        logger.info(f"  ⛓️‍💥 {step} (`{workflow_file_name}`)")
        for cve in cves:
            logger.info(
                f"      CVE: {cve.cve_id} (https://nvd.nist.gov/vuln/detail/{cve.cve_id})"
            )
            logger.info(f"      CVSSs: ")
            for cvss in cve.cvss_scores:
                logger.info(f"        - Source:{cvss.source}")
                logger.info(f"          - Base Score:{cvss.base_score}")
                logger.info(f"          - Version:{cvss.type}")
                logger.info(f"          - Severity:{cvss.severity}")
            logger.info(f"      CWEs: ")
            for weakness in cve.weaknesses:
                logger.info(
                    f"              - https://cwe.mitre.org/data/definitions/{weakness.split('CWE-')[1]}.html"
                )
            logger.info(
                f"      Description: {cve.description} (https://nvd.nist.gov/vuln/detail/{cve.cve_id})"
            )
            logger.info("\n")


def _get_steps(
    workflows: list[workflowparser.Workflow],
) -> list[Tuple[workflowparser.Step, str]]:
    steps_data = []
    for workflow in workflows:
        for job in workflow.jobs:
            for step in job.steps:
                steps_data.append((step, workflow.file_name))
    return steps_data



import logging

from httpclients.base_httpclient import BaseHttpClient
from httpclients.responsemodels.cves_response import CvesResponse

logger = logging.getLogger(__name__)


class NvdHttpClient(BaseHttpClient):
    """Derived HttpClient to handle interacting with NIST NVD (National Vulnerability Database)"""

    def __init__(self, base_url: str):
        logger.debug("initialising NvdHttpClient")
        super().__init__(base_url)
        self.cves_url = "rest/json/cves/2.0"
        self._cves_cache = {}

    def find_cves_by_keywords(self, keywords: str) -> CvesResponse:
        logger.debug("called find_cves_by_keywords")
        if keywords in self._cves_cache:
            logger.debug(f"returning cached response for keywords: {keywords}")
            return self._cves_cache[keywords]

        keywords = keywords
        logger.debug(f"calling NVD API with {keywords}")
        cves_response_json = self.get(self.cves_url, params={"keywordSearch": keywords})
        logger.debug(f"got response from NVD API\n{cves_response_json}")
        cves_response = CvesResponse.parse_json(cves_response_json)
        self._cves_cache[keywords] = cves_response
        return cves_response
